from opentab.utility.constants import OPENTAB
from opentab.utility.get_right_urls import get_right_urls
from opentab.exception.GroupExists import GroupExists

# add_tab adds the URLs contained in urls to the group group_name
# in the dictionary dic, representing the parsed tabs.yml file.
# It returns the modified dictionary.
def add_tab(group_name, dic, urls):
    if group_name in dic[OPENTAB] and urls == []:
        raise GroupExists

    # if there is no such group, create it
    if group_name not in dic[OPENTAB]:
        dic[OPENTAB][group_name] = []

    right_urls = get_right_urls(group_name, urls, dic)

    for url in right_urls:
        dic[OPENTAB][group_name].append(url)

    return dic


import argparse

# parse_cli_args parses the CLI arguments inserted by the user.
def parse_cli_args():
    parser = argparse.ArgumentParser(prog='opentab', description='Parse opentab arguments')
    # Subparsers is a subparser constructor. 
    # dest parameter is necessary to know the name of the subparser invoked by CLI.
    subparsers = parser.add_subparsers(dest='subparser')

    # Create the init subparser.
    # It is intended to be a command without arguments for initializing 
    # the .opentab folder when it is first called.
    subparsers.add_parser('init')

    # Create the reset subparser.
    # It is intended to be a command without arguments for resetting 
    # the .opentab folder whenever it is called.
    subparsers.add_parser('reset')

    # Create the add subparser with its arguments.
    # It can be specified either a group_name with no URLs
    # or a group_name and a list of one or more URLs.
    add_parser = subparsers.add_parser('add')
    add_parser.add_argument('group_name', help='Name of the group where the URL will be '
                                               + 'added to. If the group does not exist, it creates it.')
    add_parser.add_argument('urls', nargs='*', help='URL to be added to the '
                                                    + 'specified group.')

    # Create the remove subparser with its arguments
    # It can be specified either a group_name with no URLs
    # or a group_name and a list of one or more URLs.
    remove_parser = subparsers.add_parser('rm')
    remove_parser.add_argument('group_name', help='Name of the group '
                                                  + 'where the URL will be added to.')
    remove_parser.add_argument('urls', nargs='*', help='URL to be removed from the '
                                                       + 'specified group.')

    # Create the list subparser with its arguments
    # It can be issued just an 'ls' command or an 'ls' command followed by a group_name.
    list_parser = subparsers.add_parser('ls')
    list_parser.add_argument('group_name', nargs='?', default='', help='Name of the '
                                                                       + 'group to show the URLs list of.')
    list_parser.add_argument('-a', '--all', action='store_true', help='Print every group '
                                                                      + 'each with the URLs in it.')

    # Create the open subparser with its arguments.
    # It must be passed a group name for taking the corresponding URL(s).
    open_parser = subparsers.add_parser('open')
    open_parser.add_argument('group_name', help='Name of the group from which '
                                                + 'the URL will be opened.')
    open_parser.add_argument('--new-session', '-n', action='store_true',
                             help='Open a new browser session instead of opening the tabs'
                                  + 'in the same existing session.')

    # Parse the arguments.
    return parser.parse_args()


# TABS_FILE_PATH is the path of the yml file containing the tabs
# for the groups inserted by the user.
from os.path import expanduser, exists
HOME = expanduser("~")
OPENTAB_DIR_PATH = HOME + '/.opentab'
if exists(OPENTAB_DIR_PATH+'/devel'):
    TABS_FILE_PATH = OPENTAB_DIR_PATH+'/devel/tabs.yml' 
else: 
    TABS_FILE_PATH = OPENTAB_DIR_PATH+'/tabs.yml' 

# OPENTAB is the name of the first level key of the tabs.yaml file.
OPENTAB = 'opentab'

###### COLORS ######

# Define opentab, group and tabs colors.
from colorama import Fore, Style
OPENTAB_COLOR = Style.BRIGHT + Fore.YELLOW
GROUP_COLOR = Style.BRIGHT + Fore.CYAN
URL_COLOR = Style.BRIGHT + Fore.GREEN
ERROR_COLOR = Style.BRIGHT + Fore.RED
WARNING_COLOR = Style.BRIGHT + Fore.LIGHTWHITE_EX
RESET_COLOR = Style.RESET_ALL

###### URL LISTING CHAR ######

# Define chars for tabs listing.
# prefix components:
SPACE =  '    '
BRANCH = '│   '
# pointers:
TEE =    '├── '
LAST =   '└── '


from ruamel.yaml import YAML
from opentab.utility.constants import TABS_FILE_PATH

# read_yaml reads the tabs.yml file and parses it.
# It returns a dictionary parsed from the file.
def read_yaml():
    yaml = YAML()
    try:
        with open(TABS_FILE_PATH, 'r') as groups:
            dic = yaml.load(groups)
    except FileNotFoundError:
        raise FileNotFoundError
    return dic


from opentab.utility.constants import OPENTAB, ERROR_COLOR

# remove_group removes the group group_name along with its tabs from
# the dictionary dic, that represents the parsed tabs.yaml file.
def remove_group(group_name, dic):
    # For check if group exists
    try:
        del dic[OPENTAB][group_name]
    except KeyError:
        raise KeyError

    return dic


from opentab.utility.constants import GROUP_COLOR, LAST, OPENTAB, RESET_COLOR, TEE, URL_COLOR
from opentab.utility.get_group_urls import get_group_urls

# list_group_tabs lists the tabs inside the group group_name.
# To accomplish the task, it takes the dictionary dic, 
# that represents the parsed tabs.yml file.
def list_group_tabs(group_name, dic):
    try:
        size = len(dic[OPENTAB][group_name])
    except KeyError:
        raise KeyError
    i = 0
    char = TEE

    print('Tabs in ' + GROUP_COLOR + group_name + RESET_COLOR + ' are:')
    print()
    urls = get_group_urls(group_name=group_name, dic=dic)
    if urls == []:
        print('There is no tabs for the group ' + GROUP_COLOR + group_name)
        return
    print(GROUP_COLOR + group_name)
    for url in urls:
        if i == (size-1):
            char = LAST
        i += 1
        print(char + URL_COLOR + url)


from opentab.exception.GroupExists import GroupExists
from opentab.command.list_tree import list_tree
from opentab.command.reset_opentab import reset_opentab
from opentab.parsing.parse_cli_args import parse_cli_args

from opentab.command.add_tab import add_tab
from opentab.command.remove_group import remove_group
from opentab.command.remove_tabs import remove_tabs
from opentab.command.list_group_tabs import list_group_tabs
from opentab.command.list_groups import list_groups
from opentab.command.open_tabs import open_tabs
from opentab.command.init_opentab import init_opentab

from opentab.utility.constants import TABS_FILE_PATH, ERROR_COLOR, OPENTAB_COLOR, OPENTAB

from opentab.utility.get_group_urls import get_group_urls

from opentab.yaml.read_yaml import read_yaml
from opentab.yaml.write_yaml import write_yaml

from colorama import init

from os.path import exists

# Command names definition.
ADD = 'add'
RM = 'rm'
LS = 'ls'
OPEN = 'open'
CONFIG = 'config'
INIT = 'init'
RESET = 'reset'


def opentab():
    # set colorama to autoreset the colors after each print
    init(autoreset=True)

    # parse arguments provided by the user
    args = parse_cli_args()

    # get the command the user has provided
    command = args.subparser

    # the user has issued an INIT command.
    if command == INIT:
        # check whether the tabs.yml file exists in the home directory.
        # Otherwise create it. Even though the command should be issued 
        # only when initializing the opentab workspace, it is necessary
        # to anticipate any usage of the command itself.
        if not exists(TABS_FILE_PATH):
            init_opentab()
        else:
            print(ERROR_COLOR + 'The opentab folder already exists!\n')
            print('If you want to reset, '
                  + 'for any reason, the folder, run:\n')
            print(OPENTAB_COLOR + 'opentab reset\n')
            print('However, you should be careful with it as it would lead to a'
                  ' permanent loss of your saved groups and tabs.')

    # the user has issued a RESET command.
    if command == RESET:
        print('Are you sure you want to reset your tabs.yml file? [yes/n]\n')
        print(ERROR_COLOR+'Please note that this would delete ' 
            +'your saved groups and urls.')
        if input().lower() == 'yes':
            reset_opentab()
            print('Resetting...')
        else:
            print('Reset procedure canceled.')

    # read the tabs.yml file
    try:
        dic = read_yaml()
    except FileNotFoundError:
        exit(ERROR_COLOR + "File not found!!\nYou need to opentab reset")
        
    # the user has issued an ADD command.
    if command == ADD:
        group_name = args.group_name
        urls = args.urls
        try:
            dic = add_tab(group_name=group_name, dic=dic, urls=urls)
        except GroupExists:
            exit(ERROR_COLOR + "The group already exists")
        write_yaml(dic, 'w')

    # the user has issued a RM command.
    if command == RM:
        group_name = args.group_name
        urls = args.urls
        if not urls:
            try:
                dic = remove_group(group_name=group_name, dic=dic)
            except KeyError:
                exit(ERROR_COLOR + "The Group does not exists")

            write_yaml(dic, 'w')
        else:
            try:
                remove_tabs(group_name=group_name, dic=dic, urls=urls)
            except KeyError:
                exit(ERROR_COLOR + "The Group does not exists")
            write_yaml(dic, 'w')

    # the user has issued a LS command.
    if command == LS:
        group_name = args.group_name

        if group_name == '' and not args.all:
            list_groups(dic)
        elif group_name != '' and not args.all:
            try:
                list_group_tabs(group_name, dic)
            except KeyError:
                exit(ERROR_COLOR + "The Group does not exists")
        elif group_name == '' and args.all:
            list_tree(dic)

    # the user has issued an OPEN command.
    if command == OPEN:
        group_name = args.group_name

        try:
            urls_to_open = get_group_urls(group_name, dic)
        except KeyError:
            exit(ERROR_COLOR + "THE GROUP DOES NOT EXIST")

        if not urls_to_open:
            exit(ERROR_COLOR + "THERE ARE NOT URLS TO OPEN")

        if args.new_session:
            session_type = 1
        else:
            session_type = 0

        open_tabs(group_name=group_name, session_type=session_type, urls=urls_to_open)


from opentab.utility.constants import ERROR_COLOR
from opentab.utility.get_group_urls import get_group_urls
from opentab.utility.is_url_correct import is_url_correct

# get_right_urls takes the new URLs, urls, to be added to the group group_name.
# The URLs will be added to dic, the YAML file parsed dictionary.
# It returns a list of the new correct URLs.
def get_right_urls(group, urls, dic):
    # get the group urls.
    try:
        group_urls = get_group_urls(group, dic)
    except KeyError:
        exit(ERROR_COLOR + "THE GROUP DOES NOT EXIST")
    new_urls = []

    for url in urls:
        # if the inserting url does not match an established pattern
        # or it is already in the group, do not insert it.
        if is_url_correct(url) and url not in group_urls:
            new_urls.append(url)
        else:
            print(url + ' has not been recognized as a correct URL '
                + 'or it is already in the group.')

    return new_urls


from os import makedirs

from opentab.yaml.write_yaml import write_yaml
from opentab.utility.constants import TABS_FILE_PATH, OPENTAB_DIR_PATH, OPENTAB, ERROR_COLOR

# init_opentab inits the opentab workspace.
# It creates the .opentab directory and initializes the tabs.yml file
# with a first level key.
def init_opentab():
    try:
        makedirs(OPENTAB_DIR_PATH)
    except FileExistsError as fileExistsError:
        exit(ERROR_COLOR+"\nDirectory already exists, you need to opentab Reset\n")
    open(TABS_FILE_PATH, 'x')
    init_dic = {}       
    init_dic[OPENTAB] = {}
    write_yaml(init_dic, 'w')


from opentab.utility.constants import BRANCH, LAST, OPENTAB, OPENTAB_COLOR, GROUP_COLOR, SPACE, TEE, URL_COLOR

# list_tree prints the groups, along with the inner URLs, in a treeish format.
def list_tree(dic):
    group_size = len(dic[OPENTAB])
    group_i = 0
    group_char = TEE
    
    print(OPENTAB_COLOR + OPENTAB)

    # First-level elements scope.
    # Print each element along with its subelement.
    for group in dic[OPENTAB]:
        tabs_size = len(dic[OPENTAB][group])
        tabs_i = 0
        tabs_char = TEE
        branch_char = BRANCH

        # determine whether it is the last element
        if group_i == (group_size-1):
            group_char = LAST    
        print(group_char + GROUP_COLOR + group)

        # Second-level elements scope.
        # Print each subelement.
        for url in dic[OPENTAB][group]:

            # determine whether it is the last subelement
            if tabs_i == (tabs_size-1):
                tabs_char = LAST

            # if it is the last element, do not print the BRANCH symbol
            if group_char == LAST:
                branch_char = SPACE

            print(branch_char + tabs_char + URL_COLOR + url)
            tabs_i += 1

        group_i += 1


from webbrowser import open as web_open

from opentab.utility.constants import GROUP_COLOR, URL_COLOR

OPEN_IN_SAME_SESSION = 2

# open_tabs opens a group of tabs taken from group_name.
# If session_type is 0, the tabs are opened in any existing browser session. 
# If session_type is 1, a new browser session is opened. 
# If session_type is 2, new tabs will be opened next to the existing ones.
# urls are the URLs for the group group_name. 
def open_tabs(group_name, session_type, urls):
    in_session = False
    print('Opening tabs group: ' + GROUP_COLOR + group_name)
    for url in urls:
        web_open(url, new=session_type, autoraise=True)
        print('Successfully opened: ' + URL_COLOR + url)
        if not in_session:
            session_type = OPEN_IN_SAME_SESSION
            in_session = True


from opentab.utility.constants import GROUP_COLOR, LAST, OPENTAB, OPENTAB_COLOR, TEE

# list_groups prints the groups stored in opentab.
def list_groups(dic):
    size = len(dic[OPENTAB])
    i = 0
    char = TEE

    print('Groups saved in ' + OPENTAB_COLOR + 'opentab:')
    print()
    print(OPENTAB_COLOR + OPENTAB)
    for group in dic[OPENTAB]:
        if i == (size-1):
            char = LAST    
        i += 1
        print(char + GROUP_COLOR + group)


from opentab.utility.constants import TABS_FILE_PATH, OPENTAB
from opentab.command.init_opentab import init_opentab
from opentab.yaml.write_yaml import write_yaml

# reset_opentab resets the .opentab directory.
def reset_opentab():
    try:
        open(TABS_FILE_PATH, 'w')
    except:
        init_opentab()
    init_dic = {}
    init_dic[OPENTAB] = {}
    write_yaml(init_dic, 'w')


from opentab.utility.constants import OPENTAB
from opentab.yaml.read_yaml import read_yaml

# get_group_urls gets the URLs for a group group_name. If group does not exist 
# it returns an empty list, otherwise it returns the URLs in that group.
def get_group_urls(group_name, dic):
    group_urls = []
    try:
        group_urls = dic[OPENTAB][group_name]
    except KeyError:
        raise KeyError
    return group_urls


from ruamel.yaml import YAML
from opentab.utility.constants import TABS_FILE_PATH

# write_yaml writes a dictionary, dic, inside the tabs.yml file 
# respecting the YAML format specification.
# The opening mode for the file can be specified in mode parameter. 
def write_yaml(dic, mode):
    yaml = YAML()
    # mapping is the number of spaces between two YAML keys.
    # sequence is the number of spaces between the dash and the YAML value. 
    # offset is the number of spaces between the key and the dash.
    yaml.indent(mapping=2, sequence=1, offset=2)
    with open(TABS_FILE_PATH, mode) as writing_file:
        yaml.dump(dic, writing_file)



from opentab.utility.constants import OPENTAB, ERROR_COLOR

# remove_tabs removes the URLs specified in urls for the group group_name
# from dic, a dictionary resulting from the tabs.yaml parsing operation.
# It returns the modified dictionary.
def remove_tabs(group_name, dic, urls):
    print('Removing tabs for group: ' + group_name)
    for url in urls:
        try:
            if url in dic[OPENTAB][group_name]:
                dic[OPENTAB][group_name].remove(url)
            else:
                exit(ERROR_COLOR + "URL: " + url + " is not present")
        except KeyError:
            raise KeyError
    return dic